#include "FNVModel.hpp"

FNVModel::FNVModel()
{
    rMaterial = NotAssigned;

    rViscosity = Panda;

    IsViscosityNonlinear = false;
    IsMullinEffectPresent = false;
    q_nvp = 0.0;
    M_mullin = 1.0;
}






void FNVModel::SetNeoHookeanMaterialParams(double rC1e, double rC1oe, double rEta_viscosity )
{

    rMaterial = NeoHookean;

    C1e_NH  = rC1e;
    C1oe_NH = rC1oe;
    Eta_NH  = rEta_viscosity;

    if (C1e_NH <= 0.0 || C1oe_NH < 0.0 || Eta_NH <= 0.0)
    {
        std::cerr << "Error: Invalid NeoHookean material parameters !!! "  << std::endl;
        exit(-1);
    }
}
void FNVModel::SetNeoHookeanMaterialParams(double rC1e, double rC1oe, std::vector<double> rEta_viscosity )
{
    rMaterial = NeoHookean;

    C1e_NH  = rC1e;
    C1oe_NH = rC1oe;

    IsViscosityNonlinear = true;
    mu_nvp      = rEta_viscosity[0];
    xi_nvp      = rEta_viscosity[1];
    tau_max_nvp = rEta_viscosity[2];
    tau_min_nvp = rEta_viscosity[3];
    tau_q_nvp   = rEta_viscosity[4];
    zeta_nvp    = rEta_viscosity[5];

    if (C1e_NH <= 0.0 || C1oe_NH < 0.0 || mu_nvp <= 0.0)
    {
        std::cerr << "Error: Invalid NeoHookean material parameters !!! "  << std::endl;
        exit(-1);
    }
}
void FNVModel::SetNeoHookeanMaterialParams(double rC1e, double rC1oe, std::vector<double> rEta_viscosity,
                                            std::vector<double> rMullinParams )
{
    rMaterial = NeoHookean;

    C1e_NH  = rC1e;
    C1oe_NH = rC1oe;

    IsViscosityNonlinear = true;
    mu_nvp      = rEta_viscosity[0];
    xi_nvp      = rEta_viscosity[1];
    tau_max_nvp = rEta_viscosity[2];
    tau_min_nvp = rEta_viscosity[3];
    tau_q_nvp   = rEta_viscosity[4];
    zeta_nvp    = rEta_viscosity[5];

    IsMullinEffectPresent = true;
    alpha_mullin = rMullinParams[0];
    beta_mullin  = rMullinParams[1];
    gamma_mullin = rMullinParams[2];

    if (C1e_NH <= 0.0 || C1oe_NH < 0.0 || mu_nvp <= 0.0)
    {
        std::cerr << "Error: Invalid NeoHookean material parameters !!! "  << std::endl;
        exit(-1);
    }
}








void FNVModel::SetHumphreyMaterialParams(double rC1e, double rC2e, double rC1oe, double rC2oe,
                                    double rEta_viscosity )
{
    rMaterial = Humphrey;

    C1e_HM  = rC1e;
    C2e_HM  = rC2e;
    C1oe_HM = rC1oe;
    C2oe_HM = rC2oe;
    Eta_HM  = rEta_viscosity;

    if (C1e_HM <= 0.0 || C2e_HM <= 0.0 || C1oe_HM < 0.0  || C2oe_HM != C2e_HM || Eta_HM <= 0.0)
    {
        std::cerr << "Error: Invalid Humphrey material parameters !!! "  << std::endl;
        exit(-1);
    }

}
void FNVModel::SetHumphreyMaterialParams(double rC1e, double rC2e, double rC1oe, double rC2oe,
                                    std::vector<double> rEta_viscosity )
{
    rMaterial = Humphrey;

    C1e_HM  = rC1e;
    C2e_HM  = rC2e;
    C1oe_HM = rC1oe;
    C2oe_HM = rC2oe;

    IsViscosityNonlinear = true;
    mu_nvp      = rEta_viscosity[0];
    xi_nvp      = rEta_viscosity[1];
    tau_max_nvp = rEta_viscosity[2];
    tau_min_nvp = rEta_viscosity[3];
    tau_q_nvp   = rEta_viscosity[4];
    zeta_nvp    = rEta_viscosity[5];

    if (C1e_HM <= 0.0 || C2e_HM <= 0.0 || C1oe_HM < 0.0  || C2oe_HM != C2e_HM || mu_nvp <= 0.0)
    {
        std::cerr << "Error: Invalid Humphrey material parameters !!! "  << std::endl;
        exit(-1);
    }
}
void FNVModel::SetHumphreyMaterialParams(double rC1e, double rC2e, double rC1oe, double rC2oe,
                                    std::vector<double> rEta_viscosity, std::vector<double> rMullinParams )
{
    rMaterial = Humphrey;

    C1e_HM  = rC1e;
    C2e_HM  = rC2e;
    C1oe_HM = rC1oe;
    C2oe_HM = rC2oe;

    IsViscosityNonlinear = true;
    mu_nvp      = rEta_viscosity[0];
    xi_nvp      = rEta_viscosity[1];
    tau_max_nvp = rEta_viscosity[2];
    tau_min_nvp = rEta_viscosity[3];
    tau_q_nvp   = rEta_viscosity[4];
    zeta_nvp    = rEta_viscosity[5];

    IsMullinEffectPresent = true;
    alpha_mullin = rMullinParams[0];
    beta_mullin  = rMullinParams[1];
    gamma_mullin = rMullinParams[2];


    if (C1e_HM <= 0.0 || C2e_HM <= 0.0 || C1oe_HM < 0.0  || C2oe_HM != C2e_HM || mu_nvp <= 0.0)
    {
        std::cerr << "Error: Invalid Humphrey material parameters !!! "  << std::endl;
        exit(-1);
    }
}
void FNVModel::SetHumphreyMaterialParams(double rC1e, double rC2e, double rC1oe, double rC2oe,
                                    double rEta_viscosity, std::vector<double> rMullinParams )
{
    rMaterial = Humphrey;

    C1e_HM  = rC1e;
    C2e_HM  = rC2e;
    C1oe_HM = rC1oe;
    C2oe_HM = rC2oe;
    Eta_HM  = rEta_viscosity;

    IsViscosityNonlinear = false;

    IsMullinEffectPresent = true;
    alpha_mullin = rMullinParams[0];
    beta_mullin  = rMullinParams[1];
    gamma_mullin = rMullinParams[2];

    if (C1e_HM <= 0.0 || C2e_HM <= 0.0 || C1oe_HM < 0.0  || C2oe_HM != C2e_HM || Eta_HM <= 0.0)
    {
        std::cerr << "Error: Invalid Humphrey material parameters !!! "  << std::endl;
        exit(-1);
    }

}




void FNVModel::SetYeohMaterialParams(double rC1e, double rC2e, double rC1oe, double rC2oe,
                                    double rEta_viscosity )
{
    rMaterial = Yeoh;

    C1e_YM  = rC1e;
    C2e_YM  = rC2e;
    C1oe_YM = rC1oe;
    C2oe_YM = rC2oe;
    Eta_YM  = rEta_viscosity;

    if (C1e_YM <= 0.0 || C1oe_HM < 0.0   || Eta_YM <= 0.0)
    {
        std::cerr << "Error: Invalid Yeoh material parameters !!! "  << std::endl;
        exit(-1);
    }

}
void FNVModel::SetYeohMaterialParams(double rC1e, double rC2e, double rC1oe, double rC2oe,
                            std::vector<double> rEta_viscosity )
{
    rMaterial = Yeoh;

    C1e_YM  = rC1e;
    C2e_YM  = rC2e;
    C1oe_YM = rC1oe;
    C2oe_YM = rC2oe;

    IsViscosityNonlinear = true;
    mu_nvp      = rEta_viscosity[0];
    xi_nvp      = rEta_viscosity[1];
    tau_max_nvp = rEta_viscosity[2];
    tau_min_nvp = rEta_viscosity[3];
    tau_q_nvp   = rEta_viscosity[4];
    zeta_nvp    = rEta_viscosity[5];

    if (C1e_YM <= 0.0 || C1oe_HM < 0.0   || mu_nvp <= 0.0)
    {
        std::cerr << "Error: Invalid Yeoh material parameters !!! "  << std::endl;
        exit(-1);
    }
}
void FNVModel::SetYeohMaterialParams(double rC1e, double rC2e, double rC1oe, double rC2oe,
                                    std::vector<double> rEta_viscosity, std::vector<double> rMullinParams )
{
    rMaterial = Yeoh;

    C1e_YM  = rC1e;
    C2e_YM  = rC2e;
    C1oe_YM = rC1oe;
    C2oe_YM = rC2oe;

    IsViscosityNonlinear = true;
    mu_nvp      = rEta_viscosity[0];
    xi_nvp      = rEta_viscosity[1];
    tau_max_nvp = rEta_viscosity[2];
    tau_min_nvp = rEta_viscosity[3];
    tau_q_nvp   = rEta_viscosity[4];
    zeta_nvp    = rEta_viscosity[5];

    IsMullinEffectPresent = true;
    alpha_mullin = rMullinParams[0];
    beta_mullin  = rMullinParams[1];
    gamma_mullin = rMullinParams[2];

    if (C1e_YM <= 0.0 || C1oe_HM < 0.0   || mu_nvp <= 0.0)
    {
        std::cerr << "Error: Invalid Yeoh material parameters !!! "  << std::endl;
        exit(-1);
    }
}














void FNVModel::CalculateStress( std::vector < std::vector<double> > & F,
                                std::vector < std::vector<double> > & F_old,
                                std::vector < std::vector<double> > & Soe,
                                std::vector < std::vector<double> > & Be,
                                double dt ,
                                std::vector < std::vector<double> > & ReturnStress,
                                StressType rStressType,
                                LoadingType rLoadingType)
{
    //===========================================================================================
    // Class initialisation
    HelperFunction HObj;


    //===========================================================================================
    // variable assign
    if (rMaterial == NotAssigned)
    {
        std::cerr << "Error: No material law has been provided !!! "  << std::endl;
        exit(-1);
    }

    double eta_viscosity;
    if (IsViscosityNonlinear == false)
    {
        if (rMaterial == NeoHookean)
        {
            eta_viscosity = Eta_NH;
        }
        else if (rMaterial == Humphrey)
        {
            eta_viscosity = Eta_HM;
        }
        else if (rMaterial == Yeoh)
        {
            eta_viscosity = Eta_YM;
        }
        else
        {
            std::cerr << "Error: No material law has been provided !!! "  << std::endl;
            exit(-1);
        }
    }



    std::vector< std::vector<double> > Se (3,std::vector<double> (3,0.0));
    std::vector< std::vector<double> > S_Cauchy(3,std::vector<double> (3,0.0));
    std::vector< std::vector<double> > ZeroTensor (3,std::vector<double> (3,0.0));


    //===========================================================================================
    // calculation
    std::vector < std::vector<double> > InvF        = HObj.Inverse(F);
    std::vector < std::vector<double> > InvtransF   = HObj.Transpose(InvF);
    std::vector < std::vector<double> > transF      = HObj.Transpose(F);
    std::vector < std::vector<double> > B           = HObj.Multiply(F,transF);
//    std::vector < std::vector<double> > C           = HObj.Multiply(transF,F);
    std::vector < std::vector<double> > Fdot        = HObj.Subtract(F,F_old);
    Fdot= HObj.Multiply((1.0/dt),Fdot); // Fdot = (F-F_old)/dt;
    std::vector < std::vector<double> > L           = HObj.Multiply(Fdot,InvF); // L= Fdot*InvF;
    std::vector < std::vector<double> > transL      = HObj.Transpose(L);

    if (IsViscosityNonlinear == true)
    {
        if (rViscosity == Panda)
        {
            std::vector < std::vector<double> > D = HObj.Add(L,transL);
            D = HObj.Multiply(0.5,D);
            double magB = HObj.Contraction(B,B);magB = sqrt(magB);
            double magD = HObj.Contraction(D,D);magD = sqrt(magD);

            q_nvp = q_nvp + dt*(zeta_nvp*magD*(1.0-q_nvp)-(q_nvp/tau_q_nvp));
            double tau_k = tau_max_nvp*(1.0-q_nvp) + tau_min_nvp*q_nvp;
            eta_viscosity = mu_nvp*pow(magB,xi_nvp)/(magD+(1.0/tau_k));
        }
        else if (rViscosity == Haupt)
        {
            std::vector < std::vector<double> > D = HObj.Add(L,transL);
            D = HObj.Multiply(0.5,D);
            double magB = HObj.Contraction(B,B);magB = sqrt(magB);
            double magD = HObj.Contraction(D,D);magD = sqrt(magD);

            q_nvp = q_nvp + dt*(zeta_nvp*magD*(1.0-q_nvp)-(q_nvp/tau_q_nvp));
            double tau_k = tau_max_nvp*(1.0-q_nvp) + tau_min_nvp*q_nvp;
            eta_viscosity = mu_nvp*magB/(xi_nvp*magD+(1.0/tau_k));
        }
        else
        {
            eta_viscosity = -1.0;
        }
    }
    assert(eta_viscosity > 0.0);
    assert( isnan(eta_viscosity) ==false );


    std::vector < std::vector<double> > II = HObj.Identity(3);


    // // // Forward Euler method
    double trcSoe =  (1.0/3.0)*HObj.Trace(Soe);
    std::vector < std::vector<double> > tDevSoe     = HObj.Multiply(trcSoe,II);
    std::vector < std::vector<double> > dev_Soe     = HObj.Subtract(Soe,tDevSoe); // dev_Soe = Soe - (1.0/3.0)*trace(Soe)*II;

    std::vector < std::vector<double> > tSlp1       = HObj.Multiply(Be, transL);
    std::vector < std::vector<double> > tSlp2       = HObj.Multiply(L,Be);
    std::vector < std::vector<double> > tSlp3       = HObj.Multiply(Be,dev_Soe);
    tSlp3 = HObj.Multiply((2.0/eta_viscosity),tSlp3);
    std::vector < std::vector<double> > tslope = HObj.Add(tSlp1,tSlp2);
    std::vector < std::vector<double> > slope = HObj.Subtract(tslope,tSlp3);//slope = Be*L'+ L*Be-(2/eta)*Be*dev_Soe;
    std::vector < std::vector<double> > tBe = HObj.Multiply(dt,slope);
    Be= HObj.Add(Be,tBe);// Be= Be+ dt*(slope);



    // // // RK 4 method
//    std::vector < std::vector<double> > temp_Be = Be;
//    std::vector < std::vector<double> > temp_Soe;
//    Calculate_Soe(temp_Be,temp_Soe);
//    std::vector < std::vector<double> > K1_RK4;
//    SolveEvolution_Be(temp_Be,temp_Soe,L,transL,eta_viscosity,K1_RK4);
//
//    std::vector < std::vector<double> > tBe = HObj.Multiply((dt/2.0),K1_RK4);
//    temp_Be= HObj.Add(Be,tBe);
//    Calculate_Soe(temp_Be,temp_Soe);
//    std::vector < std::vector<double> > K2_RK4;
//    SolveEvolution_Be(temp_Be,temp_Soe,L,transL,eta_viscosity,K2_RK4);
//
//    tBe = HObj.Multiply((dt/2.0),K2_RK4);
//    temp_Be= HObj.Add(Be,tBe);
//    Calculate_Soe(temp_Be,temp_Soe);
//    std::vector < std::vector<double> > K3_RK4;
//    SolveEvolution_Be(temp_Be,temp_Soe,L,transL,eta_viscosity,K3_RK4);
//
//    tBe = HObj.Multiply(dt,K3_RK4);
//    temp_Be= HObj.Add(Be,tBe);
//    Calculate_Soe(temp_Be,temp_Soe);
//    std::vector < std::vector<double> > K4_RK4;
//    SolveEvolution_Be(temp_Be,temp_Soe,L,transL,eta_viscosity,K4_RK4);
//
//    K2_RK4 = HObj.Multiply(2.0,K2_RK4);
//    K3_RK4 = HObj.Multiply(2.0,K3_RK4);
//    std::vector < std::vector<double> > slope = HObj.Add(K1_RK4,K2_RK4);
//    slope = HObj.Add(slope,K3_RK4);
//    slope = HObj.Add(slope,K4_RK4);
//    slope = HObj.Multiply((dt/6.0),slope);
//    Be    = HObj.Add(Be,slope);



    if (rMaterial == NeoHookean)
    {
        Soe = HObj.Multiply(2.0*C1oe_NH,Be); //
        Se  = HObj.Multiply(2.0*C1e_NH,B);    //
    }
    else if(rMaterial == Humphrey)
    {
        double I1B  = HObj.Trace(B);
        double I1Be = HObj.Trace(Be);
        Soe = HObj.Multiply(2.0*C1oe_HM*C2oe_HM*exp(C2oe_HM*(I1Be-3.0)),Be); //
        Se  = HObj.Multiply(2.0*C1e_HM*C2e_HM*exp(C2e_HM*(I1B-3.0)),B);    //
    }
    else if(rMaterial == Yeoh)
    {
        double I1B  = HObj.Trace(B);
        double I1Be = HObj.Trace(Be);
        Soe = HObj.Multiply(2.0*(C1oe_YM+2.0*(C2oe_YM*(I1Be-3.0))),Be); //
        Se  = HObj.Multiply(2.0*(C1e_YM+2.0*(C2e_YM*(I1B-3.0))),B);    //
    }
    else
    {
        Soe = ZeroTensor;
        Se  = ZeroTensor;
    }

    S_Cauchy = HObj.Add(Se,Soe);// Cauchy = Se+Soe

    if (rLoadingType == UniaxialLoading)
    {
        double pressure = Se[2][2]+Soe[2][2];/// valid only for uniaxial loading in x direction
        std::vector < std::vector<double> > tPressure = HObj.Multiply(pressure,II);
        S_Cauchy = HObj.Subtract(S_Cauchy,tPressure);// Cauchy = Se+Soe-pressure*II
    }

    //Mullins effect
    if (IsMullinEffectPresent == true)
    {
        double m_mullin = HObj.Contraction(B,B);m_mullin = sqrt(m_mullin);
        if (m_mullin > M_mullin)
        {
            M_mullin = m_mullin;
        }
        double Kappa = exp(-1.0*beta_mullin*pow(((M_mullin-m_mullin)*pow((m_mullin/M_mullin),gamma_mullin)),alpha_mullin));
        assert( isnan(Kappa) ==false );
        assert(Kappa <= 1.0);
        S_Cauchy = HObj.Multiply(Kappa,S_Cauchy);
    }




    if (rStressType == CauchyStress)
    {
        ReturnStress = S_Cauchy;
    }
    else if (rStressType == PK1Stress)
    {
//        double detF = 1.0;
        std::vector < std::vector<double> > temp_PK1 = HObj.Multiply(InvF,S_Cauchy);
//        temp_PK1 = HObj.Multiply(detF,temp_PK1);
        ReturnStress = temp_PK1;

    }
    else if (rStressType == PK2Stress)
    {
//        double detF = 1.0;
        std::vector < std::vector<double> > temp_PK2 = HObj.Multiply(InvF,S_Cauchy);
        temp_PK2 = HObj.Multiply(temp_PK2,InvtransF);
//        temp_PK2 = HObj.Multiply(detF,temp_PK2);
        ReturnStress = temp_PK2;
    }
    else
    {
        ReturnStress = ZeroTensor;
    }

    F_old = F;

}





void FNVModel::Calculate_Soe(std::vector < std::vector<double> > & Be,
                                std::vector < std::vector<double> > & Soe)
{
    HelperFunction HObj;

    std::vector< std::vector<double> > ZeroTensor (3,std::vector<double> (3,0.0));


    if (rMaterial == NeoHookean)
    {
        Soe = HObj.Multiply(2.0*C1oe_NH,Be); //
    }
    else if(rMaterial == Humphrey)
    {
        double I1Be = HObj.Trace(Be);
        Soe = HObj.Multiply(2.0*C1oe_HM*C2oe_HM*exp(C2oe_HM*(I1Be-3.0)),Be); //
    }
    else if(rMaterial == Yeoh)
    {
        double I1Be = HObj.Trace(Be);
        Soe = HObj.Multiply(2.0*(C1oe_YM+2.0*(C2oe_YM*(I1Be-3.0))),Be); //
    }
    else
    {
        Soe = ZeroTensor;
    }
}





void FNVModel::SolveEvolution_Be(std::vector < std::vector<double> > & Be,
                                 std::vector < std::vector<double> > & Soe,
                                 std::vector < std::vector<double> > & L,
                                 std::vector < std::vector<double> > & transL,
                                 double eta,
                                 std::vector < std::vector<double> > & BeDot)
{
    HelperFunction HObj;
    std::vector < std::vector<double> > II = HObj.Identity(3);

    std::vector < std::vector<double> > temp_Be = Be;
    std::vector < std::vector<double> > temp_Soe = Soe;

    double trcSoe =  (1.0/3.0)*HObj.Trace(temp_Soe);
    std::vector < std::vector<double> > tDevSoe     = HObj.Multiply(trcSoe,II);
    std::vector < std::vector<double> > dev_Soe     = HObj.Subtract(temp_Soe,tDevSoe);
    std::vector < std::vector<double> > tSlp1       = HObj.Multiply(temp_Be, transL);
    std::vector < std::vector<double> > tSlp2       = HObj.Multiply(L,temp_Be);
    std::vector < std::vector<double> > tSlp3       = HObj.Multiply(temp_Be,dev_Soe);
    tSlp3 = HObj.Multiply((2.0/eta),tSlp3);
    std::vector < std::vector<double> > tslope = HObj.Add(tSlp1,tSlp2);
    BeDot  = HObj.Subtract(tslope,tSlp3);
}




void FNVModel::SetViscosityType(ViscosityType mViscosity)
{
    rViscosity = mViscosity;
}







