#ifndef FNVMODEL_HPP
#define FNVMODEL_HPP

#include <iostream>
#include <fstream>
#include <cmath>
#include<assert.h>
#include<vector>
#include <stdlib.h>
#include <math.h>

#include "HelperFunction.hpp"



class FNVModel
{
    public:

        // Enumeration for LoadingType
        enum LoadingType: int
        {
            UniaxialLoading,
            MultiaxialLoading
        };

        // Enumeration for stress types
        enum StressType: int
        {
            PK1Stress,
            PK2Stress,
            CauchyStress
        };

        // Enumeration for material types
        enum MaterialType: int
        {
            NeoHookean,
            Humphrey,
            Yeoh,
            NotAssigned
        };

        // Enumeration for viscosity types
        enum ViscosityType: int
        {
            Haupt,
            Panda
        };


        FNVModel();


        void SetNeoHookeanMaterialParams(double rC1e, double rC1oe, double rEta_viscosity );
        void SetNeoHookeanMaterialParams(double rC1e, double rC1oe, std::vector<double> rEta_viscosity );
        void SetNeoHookeanMaterialParams(double rC1e, double rC1oe, std::vector<double> rEta_viscosity,
                                            std::vector<double> rMullinParams );


        void SetHumphreyMaterialParams(double rC1e, double rC2e, double rC1oe, double rC2oe,
                                    double rEta_viscosity );
        void SetHumphreyMaterialParams(double rC1e, double rC2e, double rC1oe, double rC2oe,
                                    std::vector<double> rEta_viscosity );
        void SetHumphreyMaterialParams(double rC1e, double rC2e, double rC1oe, double rC2oe,
                                    std::vector<double> rEta_viscosity, std::vector<double> rMullinParams );
        void SetHumphreyMaterialParams(double rC1e, double rC2e, double rC1oe, double rC2oe,
                                    double rEta_viscosity, std::vector<double> rMullinParams );


        void SetYeohMaterialParams(double rC1e, double rC2e, double rC1oe, double rC2oe,
                                    double rEta_viscosity );
        void SetYeohMaterialParams(double rC1e, double rC2e, double rC1oe, double rC2oe,
                                    std::vector<double> rEta_viscosity );
        void SetYeohMaterialParams(double rC1e, double rC2e, double rC1oe, double rC2oe,
                                    std::vector<double> rEta_viscosity, std::vector<double> rMullinParams );

        void SetViscosityType(ViscosityType mViscosity);

        void CalculateStress(std::vector < std::vector<double> > & F,
                             std::vector < std::vector<double> > & F_old,
                             std::vector < std::vector<double> > & Soe,
                             std::vector < std::vector<double> > & Be,
                             double dt,
                             std::vector < std::vector<double> > & ReturnStress,
                             StressType rStressType,
                             LoadingType rLoadingType);


    protected:
    private:

        MaterialType rMaterial;

        // neohookean material parameters
        double C1e_NH;
        double C1oe_NH;
        double Eta_NH;

        // Humphrey material parameters
        double C1e_HM;
        double C1oe_HM;
        double C2e_HM;
        double C2oe_HM;
        double Eta_HM;

        // Yeoh material parameters
        double C1e_YM;
        double C1oe_YM;
        double C2e_YM;
        double C2oe_YM;
        double Eta_YM;

        // ViscosityType
        ViscosityType rViscosity;

        // Nonlinear viscosity
        double mu_nvp;// nvp: nonlinear viscosity parameter
        double xi_nvp;
        double tau_max_nvp;
        double tau_min_nvp;
        double tau_q_nvp;
        double zeta_nvp;
        double q_nvp;
        bool IsViscosityNonlinear;

        // Mullin's effect patameters
        double alpha_mullin;
        double beta_mullin;
        double gamma_mullin;
        double M_mullin;
        bool IsMullinEffectPresent;

        void Calculate_Soe(std::vector < std::vector<double> > & Be,
                            std::vector < std::vector<double> > & Soe);

        void SolveEvolution_Be(std::vector < std::vector<double> > & Be,
                                std::vector < std::vector<double> > & Soe,
                                std::vector < std::vector<double> > & L,
                                std::vector < std::vector<double> > & transL,
                                double eta,
                                std::vector < std::vector<double> > & BeDot);




};

#endif // FNVMODEL_HPP
