#include "HelperFunction.hpp"

HelperFunction::HelperFunction()
{
    //ctor
}


void HelperFunction::LUDecomposition(std::vector< std::vector<double> > & A,
                                        std::vector< std::vector<double> > & LowerT,
                                        std::vector< std::vector<double> > & UpperT)
{
    std::vector< std::vector<double> > U = UpperT;
    std::vector< std::vector<double> > L = LowerT;

    assert(A.size() == A[0].size());// check square matrix
    assert(L.size() == L[0].size());// check square matrix
    assert(U.size() == U[0].size());// check square matrix
    assert(A.size() == U.size());// check
    assert(U.size() == L.size());// check



    for (unsigned i=0;i<U.size();i++)
    {
        for (unsigned j=0; j<U[0].size();j++)
        {
            U[i][j]=0.0;
            L[i][j]=0.0;
        }
    }


    int MatSize = A.size();
    for (int i = 0; i < MatSize; i++)
    {
        for (int j = 0; j < MatSize; j++)
        {
            if (j < i)
            {
                L[j][i] = 0;
            }
            else
            {
                L[j][i] = A[j][i];
                for (int k = 0; k < i; k++)
                {
                    L[j][i] = L[j][i] - L[j][k] * U[k][i];
                }
            }
        }

        for (int j = 0; j < MatSize; j++)
        {
            if (j < i)
            {
                U[i][j] = 0;
            }
            else if (j == i)
            {
                U[i][j] = 1;
            }
            else
            {
                U[i][j] = A[i][j] / L[i][i];
                for (int k = 0; k < i; k++)
                {
                    U[i][j] = U[i][j] - ((L[i][k] * U[k][j]) / L[i][i]);
                }
            }
        }
    }

    LowerT = L;
    UpperT = U;
}


std::vector<double> HelperFunction::SolveLinearEquations(std::vector< std::vector<double> > & A,
                                                                    std::vector<double> & B)
{
    assert(A.size() == B.size());// check
    int MatSize = A.size();

    std::vector<double> x(MatSize,0.0);
    std::vector<double> z(MatSize,0.0);

    std::vector< std::vector<double> > L (MatSize,std::vector<double> (MatSize,0));
    std::vector< std::vector<double> > U (MatSize,std::vector<double> (MatSize,0));
    LUDecomposition(A,L,U);

    //***** FINDING Z; LZ=b*********//
    for(int i=0;i<MatSize;i++)
    {                                        //forward subtitution method
        double sum=0;
        for(int p=0;p<i;p++)
        {
            sum+=L[i][p]*z[p];
        }

        z[i]=(B[i]-sum)/L[i][i];
    }
    //********** FINDING X; UX=Z***********//
    for(int i=(MatSize-1);i>=0;i--)
    {
        double sum=0;
        for(int p=(MatSize-1);p>i;p--)
        {
            sum+=U[i][p]*x[p];
        }
        x[i]=(z[i]-sum)/U[i][i];
    }

    return x;
}


double HelperFunction::Contraction (std::vector< std::vector<double> > & A,std::vector< std::vector<double> > & B )
{
    assert(A.size() == B.size());
    assert(A[0].size() == B[0].size());
    double mag=0.0;
    int row = A.size(); int col = A[0].size();
    for (int i=0;i< row ;i++)
    {
        for (int j=0;j< col ;j++)
        {
            mag = mag + A[i][j] * B[i][j];
        }
    }
    return mag;
}

std::vector< std::vector<double> > HelperFunction::Dyadic (std::vector<double> & A,std::vector<double> & B )
{
    assert(A.size() == B.size());
    std::vector< std::vector<double> > Result (A.size(), std::vector<double> (A.size(),0.0));
    for (unsigned i=0; i< A.size(); i++)
    {
        for (unsigned j=0; j< A.size(); j++)
        {
            Result[i][j] = A[i]*B[j];
        }
    }

    return Result;
}


std::vector< std::vector<double> > HelperFunction::Add (std::vector< std::vector<double> > & A,
                                                        std::vector< std::vector<double> > & B )
{
    assert(A.size() == B.size());
    assert(A[0].size() == B[0].size());
    std::vector< std::vector<double> > Result (A.size(), std::vector<double> (A[0].size(),0.0));
    int row = A.size(); int col = A[0].size();
    for (int i=0;i< row ;i++)
    {
        for (int j=0;j< col ;j++)
        {
            Result[i][j] = A[i][j] + B[i][j];
        }
    }
    return Result;
}

std::vector< std::vector<double> > HelperFunction::Add (std::vector< std::vector<double> > & A,
                                                        std::vector< std::vector<double> > & B,
                                                        std::vector< std::vector<double> > & C )
{
    assert(A.size() == B.size());
    assert(A[0].size() == B[0].size());
    assert(A.size() == C.size());
    assert(A[0].size() == C[0].size());
    std::vector< std::vector<double> > Result =  Add(A,B);
    Result = Add(Result,C);
    return Result;
}

std::vector< std::vector<double> > HelperFunction::Add (std::vector< std::vector<double> > & A,
                                                        std::vector< std::vector<double> > & B,
                                                        std::vector< std::vector<double> > & C,
                                                        std::vector< std::vector<double> > & D )
{
    assert(A.size() == B.size());
    assert(A[0].size() == B[0].size());
    assert(A.size() == C.size());
    assert(A[0].size() == C[0].size());
    assert(A.size() == D.size());
    assert(A[0].size() == D[0].size());
    std::vector< std::vector<double> > Result =  Add(A,B,C);
    Result = Add(Result,D);
    return Result;
}

std::vector<double> HelperFunction::Add (std::vector<double> & A, std::vector<double> & B )
{
    assert(A.size() == B.size());
    std::vector<double> Result (A.size(), 0.0);
    for (unsigned i=0;i< Result.size() ;i++)
    {
        Result[i] = A[i] + B[i];
    }
    return Result;
}

std::vector<double> HelperFunction::Add (std::vector<double> & A, std::vector<double> & B, std::vector<double> & C )
{
    assert(A.size() == B.size());
    assert(A.size() == C.size());
    std::vector<double> Result = Add(A,B);
    Result = Add(Result,C);
    return Result;
}


std::vector< std::vector<double> > HelperFunction::Subtract (std::vector< std::vector<double> > & A,
                                                                std::vector< std::vector<double> > & B )
{
    assert(A.size() == B.size());
    assert(A[0].size() == B[0].size());
    std::vector< std::vector<double> > Result (A.size(), std::vector<double> (A[0].size(),0.0));
    int row = A.size(); int col = A[0].size();
    for (int i=0;i< row ;i++)
    {
        for (int j=0;j< col ;j++)
        {
            Result[i][j] = A[i][j] - B[i][j];
        }
    }
    return Result;
}




std::vector<double> HelperFunction::Subtract (std::vector<double> & A, std::vector<double> & B )
{
    assert(A.size() == B.size());
    std::vector<double> Result (A.size(), 0.0);
    for (unsigned i=0;i< Result.size() ;i++)
    {
        Result[i] = A[i] - B[i];
    }
    return Result;
}



std::vector< std::vector<double> > HelperFunction::Multiply (std::vector< std::vector<double> > & A,
                                                                std::vector< std::vector<double> > & B )
{
    assert(A[0].size() == B.size());
    std::vector< std::vector<double> > Result (A.size(), std::vector<double> (B[0].size(),0.0));
    for (unsigned i= 0; i< Result.size(); i++)
    {
        for (unsigned j=0;j<Result[0].size();j++)
        {
            for(unsigned k=0; k< A[0].size(); k++)
            {
                Result[i][j]= Result[i][j] + A[i][k]*B[k][j];
            }
        }
    }
    return Result;
}

std::vector< std::vector<double> > HelperFunction::Multiply (std::vector< std::vector<double> > & A,
                                                                std::vector< std::vector<double> > & B,
                                                                std::vector< std::vector<double> > & C )
{
    assert(A[0].size() == B.size());
    assert(A[0].size() == C.size());
    std::vector< std::vector<double> > Result = Multiply(A,B);
    Result = Multiply(Result, C);

    return Result;
}


std::vector< std::vector<double> > HelperFunction::Multiply ( double A, std::vector< std::vector<double> > & B )
{
    std::vector< std::vector<double> > Result (B.size(), std::vector<double> (B[0].size(),0.0));
    for (unsigned i= 0; i< Result.size(); i++)
    {
        for (unsigned j=0;j<Result[0].size();j++)
        {
            Result[i][j]= A * B[i][j];
        }
    }
    return Result;
}



double HelperFunction::Determinant ( std::vector< std::vector<double> > & A )
{
    assert(A.size() == A[0].size());
    assert( A.size() < 4 );
    double Result;
    if (A.size()==2)
    {
        Result = A[0][0]*A[1][1] - A[0][1]*A[1][0];
    }
    else if (A.size()==3)
    {
        Result = A[0][0]*(A[1][1]*A[2][2]-A[1][2]*A[2][1])
                -A[0][1]*(A[1][0]*A[2][2]-A[2][0]*A[1][2])
                +A[0][2]*(A[1][0]*A[2][1]-A[1][1]*A[2][0]);
    }
    return Result;
}




std::vector< std::vector<double> > HelperFunction::Inverse ( std::vector< std::vector<double> > & A )
{
    assert(A.size() == A[0].size());
    std::vector< std::vector<double> > Result (A.size(), std::vector<double> (A[0].size(),0.0));
    double det = Determinant(A);
    assert(det != 0.0);
    double invdet = 1/det;

    if (A.size()==2)
    {
        Result[0][0]=(A[1][1])*invdet;
        Result[0][1]=(-A[0][1])*invdet;
        Result[1][0]=(-A[1][0])*invdet;
        Result[1][1]=(A[0][0])*invdet;
    }
    else if (A.size()==3)
    {
       Result[0][0]=(A[1][1]*A[2][2]-A[1][2]*A[2][1])*invdet;
       Result[0][1]=(A[0][2]*A[2][1]-A[0][1]*A[2][2])*invdet;
       Result[0][2]=(A[0][1]*A[1][2]-A[0][2]*A[1][1])*invdet;
       Result[1][0]=(A[1][2]*A[2][0]-A[1][0]*A[2][2])*invdet;
       Result[1][1]=(A[0][0]*A[2][2]-A[0][2]*A[2][0])*invdet;
       Result[1][2]=(A[0][2]*A[1][0]-A[0][0]*A[1][2])*invdet;
       Result[2][0]=(A[1][0]*A[2][1]-A[1][1]*A[2][0])*invdet;
       Result[2][1]=(A[0][1]*A[2][0]-A[0][0]*A[2][1])*invdet;
       Result[2][2]=(A[0][0]*A[1][1]-A[0][1]*A[1][0])*invdet;
    }
    return Result;
}



std::vector< std::vector<double> >  HelperFunction::Identity ( double A )
{
   int temp_A = A;
   std::vector< std::vector<double> > Result(A,std::vector<double> (temp_A,0.0));
   if(A==2)
   {
        Result[0][0]=1.0;Result[1][1]=1.0;
   }
   else if ( A==3)
   {
        Result[0][0]=1.0;Result[1][1]=1.0;Result[2][2]=1.0;
   }
    return Result;
}

std::vector< std::vector<double> > HelperFunction::ZeroTensor ( double A )
{
   int temp_A = A;
   std::vector< std::vector<double> > Result(A,std::vector<double> (temp_A,0.0));
   return Result;
}


double HelperFunction::Trace ( std::vector< std::vector<double> > & A )
{
    assert(A.size() == A[0].size());
    double Result;
    if (A.size() ==2)
    {
        Result = A[0][0]+A[1][1];
    }
    else if (A.size()==3)
    {
         Result = A[0][0]+A[1][1]+A[2][2];
    }

    return Result;
}

std::vector< std::vector<double> > HelperFunction::DeviatoricPartofTensor ( std::vector< std::vector<double> > & A )
{
    assert(A.size() == A[0].size());
    double tr_A =  (1.0/3.0)*Trace(A);
    std::vector < std::vector<double> > II = Identity(3);
    std::vector < std::vector<double> > tDevA     = Multiply(tr_A,II);
    std::vector < std::vector<double> > dev_A     = Subtract(A,tDevA);
    return dev_A;
}

std::vector< std::vector<double> > HelperFunction::Transpose ( std::vector< std::vector<double> > & A )
{
    assert(A.size() == A[0].size());
    std::vector< std::vector<double> > Result(A.size(), std::vector<double> (A[0].size(),0.0));
    for (unsigned i= 0; i< Result.size(); i++)
    {
        for (unsigned j=0;j<Result[0].size();j++)
        {
            Result[i][j]= A[j][i];
        }
    }
    return Result;
}


void HelperFunction::PrintTensor(std::vector< std::vector<double> > & A)
{
    std::cout << "=========================" << std::endl;
    for (unsigned i = 0;i < A.size(); i++)
    {
        for (unsigned j = 0; j < A[0].size(); j++)
        {
            std::cout << A[i][j] << "\t";
        }
        std::cout << "\n";
    }
    std::cout << "=========================" << std::endl;
}

void HelperFunction::PrintTensor(std::vector< double > & A)
{
    std::cout << "=========================" << std::endl;
    for (unsigned i=0;i<A.size();i++)
    {
        std::cout << A[i] << "\t";
    }
    std::cout << "\n" << "=========================" << std::endl;
}


long double HelperFunction::Factorial(int A)
{
    assert(A >= 0 && A < 15);
    int FactValue = 1;
    for (int i=1;i<=A; i++)
    {
        FactValue = FactValue * i;
    }
    return FactValue;
}


double HelperFunction::GetExponentialIntegralValue(double A)
{
    assert(A>0);
    double IntegralValue=0.0;
    if (A>8.0)
    {
        IntegralValue = 0.0;
    }
    else
    {
        double LowerLimit = A;
        double UpperLimit = 8.0;
        double dx;
        if (LowerLimit < 0.001)
        {
            dx = 0.00001;
        }
        else
        {
            dx = 0.001;
        }
        int NumPoints = (UpperLimit - LowerLimit)/dx;

        double PrevFunValue = 0.0;
        for (int i=0; i<= NumPoints; i++)
        {
            double x = LowerLimit + dx*i;
            double FunValue = exp(-1.0*x)/x;// exponential integral function
            if (i >0)
            {
                IntegralValue = IntegralValue + 0.5*dx*(PrevFunValue+FunValue);
            }
            PrevFunValue = FunValue;
        }
    }
    assert (IntegralValue < 14.0);// integrate exp(-x)/x from 1e-6 to infinite = 13.2383
    return IntegralValue;
}




















