#ifndef SOLVENONLINEAREQUATION_HPP
#define SOLVENONLINEAREQUATION_HPP


#include <iostream>
#include <fstream>
#include <cmath>
#include<assert.h>
#include<vector>
#include <stdlib.h>
#include <math.h>       /* isnan, sqrt */

#include "HelperFunction.hpp"
#include "FNVModel.hpp"


class SolveNonlinearEquation
{
    public:

        // Enumeration for material types
        enum MaterialType: int
        {
            NeoHookean,
            Humphrey
        };


        SolveNonlinearEquation();

        void SetInitialGuess(double rInitialGuess);
        void SetParameterValues(double rCe, double rCoe, double rlambda, double rlambda_e,
                                    double rdt, double reta, double rStress);
        double SolveEquation();



        // new for Golden section
        double FindRootGoldenSection(double DesiredPressure, double Guess, double rdt);
        void SetGeometryParamsForGoldenSection(double rA_ref, double rB_ref, double rdR_ref, double rlambda_axialstretch );
        void SetMaterialParamsForGoldenSection(double rCe_NH, double rCoe_NH, double reta_NH);
        void SetMaterialParamsForGoldenSection(double rC1e_HM, double rC2e_HM, double rC1oe_HM, double rC2oe_HM, double reta_HM);





    protected:

    private:

        double mInitialGuess;
        bool IsInitialGuessPresent;
        double FunEvaTol;
        int MaxFunEvaNumber;

        double Ce;
        double Coe;
        double lambda;
        double lambda_e;
        double dt;
        double eta;
        double Stress;

        double CalculateFunctionValue(double rGuess);
        double CalculateGradientValue(double rGuess);





        // new for Golden section
        double A_ref;
        double B_ref;
        double lambda_axialstretch;
        double dR_ref;
        int    num_step_space;

        double Ce_NH;
        double Coe_NH;
        double eta_NH;

        double C1e_HM;
        double C2e_HM;
        double C1oe_HM;
        double C2oe_HM;
        double eta_HM;
        MaterialType rMaterialType;




        bool IsGoldenSectionParamsPresent;

        std::vector< std::vector< std::vector<double> > > F_old_global ;
        std::vector< std::vector< std::vector<double> > > Be_global ;
        std::vector< std::vector< std::vector<double> > > Soe_global;


        double CalculateObjectiveFunction(double Guess, double rdt,
                                            std::vector< std::vector< std::vector<double> > > & F_old,
                                            std::vector< std::vector< std::vector<double> > > & Be,
                                            std::vector< std::vector< std::vector<double> > > & Soe,
                                            bool RewriteTensors);



};

#endif // SOLVENONLINEAREQUATION_HPP
